from pyramid_oas3.jsonschema import _utils
from pyramid_oas3.jsonschema.exceptions import ValidationError


def ref(validator, ref, instance, schema, errors):
    scope, resolved = validator.resolver.resolve(ref)
    validator.resolver.push_scope(scope)
    try:
        return validator._descend(instance, resolved, errors)
    finally:
        validator.resolver.pop_scope()


def type(validator, types, instance, schema, errors):
    types = _utils.ensure_list(types)
    if not any(validator.is_type(instance, type, schema) for type in types):
        errors.append(ValidationError('{} is not of type {}'.format(
            instance, ', '.join(types))))


def items(validator, items, instance, schema, errors):
    if not validator.is_type(instance, 'array', schema):
        return

    if validator.is_type(items, 'object', schema):
        new_instance = [None] * len(instance)
        new_schema = items
        for index, item in enumerate(instance):
            new_instance[index], new_schema = validator._descend(
                item, items, errors, path=index)
    else:
        new_instance = [None] * min(len(instance), len(items))
        new_schema = list(items)
        for (index, item), subschema in zip(enumerate(instance), items):
            new_instance[index], new_schema[index] = validator._descend(
                item, subschema, errors, path=index, schema_path=index)
    return new_instance, new_schema


def properties(validator, properties, instance, schema, errors):
    if not validator.is_type(instance, 'object', schema):
        return

    new_schema = {}
    new_instance = dict(instance)
    for property, subschema in properties.items():
        prop_value = instance.get(property)
        if prop_value is None:
            filled, default_value = _try_get_default_value(
                validator, subschema)
            if filled:
                if isinstance(default_value, (dict, list)):
                    prop_value = default_value
                else:
                    # dict/list以外の場合はformatで型が変わっている可能性があるので
                    # default値を信頼してvalidationは実施しない
                    new_instance[property] = default_value
                    new_schema[property] = subschema
                    continue
            else:
                new_schema[property] = subschema
                continue
        new_instance[property], new_schema[property] = validator._descend(
            prop_value, subschema, errors, path=property, schema_path=property)
    return new_instance, new_schema


def _try_get_default_value(validator, subschema):
    if not validator._fill_by_default or not isinstance(subschema, dict):
        return False, None
    ref_value = subschema.get('$ref')
    if ref_value:
        scope, subschema = validator.resolver.resolve(ref_value)
        validator.resolver.push_scope(scope)
    try:
        if 'default' in subschema:
            return True, subschema['default']
        if 'allOf' in subschema:
            ret = []
            for s in subschema['allOf']:
                filled, value = _try_get_default_value(validator, s)
                if filled:
                    ret.append(value)
            if len(ret) == 1:
                return True, ret[0]
    finally:
        if ref_value:
            validator.resolver.pop_scope()
    return False, None


def format(validator, format, instance, schema, errors):
    if validator.format_checker is not None:
        try:
            instance = validator.format_checker.check(instance, format)
        except Exception as e:
            errors.append(ValidationError(str(e), cause=e.__cause__))
    return instance, format


def oneOf(validator, oneOf, instance, schema, errors):
    valids = []
    all_errors = []
    for index, subschema in enumerate(oneOf):
        erroff = len(errors)
        ret = validator._descend(
            instance, subschema, errors, schema_path=index)
        if len(errors) == erroff:
            valids.append((ret, subschema))
        else:
            all_errors.extend(errors[erroff:])
            del errors[erroff:]

    if len(valids) == 1:
        return valids[0][0][0], oneOf

    if not valids:
        errors.append(ValidationError(
            '{} is not valid under any of the given schemas'.format(instance),
            context=all_errors,
        ))
    elif len(valids) > 1:
        errors.append(ValidationError(
            '{} is valid under each of {}'.format(
                instance, ', '.join(repr(schema) for _, schema in valids))))


def allOf(validator, allOf, instance, schema, errors):
    if not allOf:
        return
    ret = []
    for index, subschema in enumerate(allOf):
        tmp, _ = validator._descend(
            instance, subschema, errors, schema_path=index)
        ret.append(tmp)
    return _utils.merge_instances(ret), allOf


def anyOf(validator, anyOf, instance, schema, errors):
    valids, all_errors = [], []
    for index, subschema in enumerate(anyOf):
        erroff = len(errors)
        ret, _ = validator._descend(
            instance, subschema, errors, schema_path=index)
        if len(errors) == erroff:
            valids.append(ret)
        else:
            all_errors.extend(errors[erroff:])
            del errors[erroff:]

    if len(valids) > 0:
        return _utils.merge_instances(valids), anyOf

    errors.append(ValidationError(
        '{} is not valid under any of the given schemas'.format(instance),
        context=all_errors,
    ))


def required(validator, required, instance, schema, errors):
    if not validator.is_type(instance, 'object', schema):
        return
    for property in required:
        if property not in instance:
            errors.append(ValidationError(
                '{} is a required property'.format(property)))


def additionalItems(validator, aI, instance, schema, errors):
    if (not validator.is_type(instance, 'array', schema) or
            validator.is_type(schema.get('items', {}), 'object', schema)):
        return

    items_array = schema.get('items', [])
    if validator.is_type(aI, 'object', schema):
        new_instance = list(instance)
        for index, item in enumerate(
                instance[len(items_array):], start=len(items_array)):
            new_instance[index] = validator._descend(
                item, aI, errors, path=index)
    elif not aI and len(instance) > len(items_array):
        errors.append(ValidationError(
            'Additional items are not allowed ({} were unexpected)'.format(
                ', '.join([
                    str(x) for x in instance[len(items_array):]]))))


def additionalProperties(validator, aP, instance, schema, errors):
    if not validator.is_type(instance, 'object', schema):
        return

    extras = set()
    properties = schema.get("properties", {})
    patterns = schema.get("patternProperties", {})
    if patterns and isinstance(patterns, dict):
        patterns = validator._get_regex('|'.join(patterns))
    else:
        patterns = None
    for name in instance:
        if name not in properties:
            if patterns and patterns.search(name):
                continue
            extras.add(name)

    if validator.is_type(aP, 'object', schema):
        new_instance = dict(instance)
        for extra in extras:
            new_instance[extra], _ = validator._descend(
                instance[extra], aP, errors, path=extra)
        return new_instance, aP
    elif not aP and extras:
        if patterns:
            patterns = sorted(schema['patternProperties'])
            errors.append(ValidationError(
                '{} do not match any of the regexes: {}'.format(
                    ', '.join(map(repr, sorted(extras))),
                    ', '.join(map(repr, patterns)))))
        else:
            errors.append(ValidationError(
                'Additional properties are not allowed '
                '({} were unexpected)'.format(extras)))


def minItems(validator, minItems, instance, schema, errors):
    if (validator.is_type(instance, 'array', schema)
            and len(instance) < minItems):
        errors.append(ValidationError('{} is too short'.format(instance,)))


def maxItems(validator, maxItems, instance, schema, errors):
    if (validator.is_type(instance, 'array', schema)
            and len(instance) > maxItems):
        errors.append(ValidationError('{} is too long'.format(instance,)))


def minLength(validator, mL, instance, schema, errors):
    if (validator.is_type(instance, 'string', schema)
            and len(instance) < mL):
        errors.append(ValidationError('{} is too short'.format(instance,)))


def maxLength(validator, mL, instance, schema, errors):
    if (validator.is_type(instance, 'string', schema)
            and len(instance) > mL):
        errors.append(ValidationError('{} is too long'.format(instance,)))


def minimum(validator, minimum, instance, schema, errors):
    if not validator.is_type(instance, 'number', schema):
        return

    if schema.get('exclusiveMinimum', False):
        failed = instance <= minimum
        cmp = 'less than or equal to'
    else:
        failed = instance < minimum
        cmp = 'less than'
    if failed:
        errors.append(ValidationError(
            '{} is {} the minimum of {}'.format(instance, cmp, minimum)))


def maximum(validator, maximum, instance, schema, errors):
    if not validator.is_type(instance, 'number', schema):
        return

    if schema.get('exclusiveMaximum', False):
        failed = instance >= maximum
        cmp = 'greater than or equal to'
    else:
        failed = instance > maximum
        cmp = 'greater than'

    if failed:
        errors.append(ValidationError(
            '{} is {} the maximum of {}'.format(instance, cmp, maximum)))


def minProperties(validator, mP, instance, schema, errors):
    if validator.is_type(instance, 'object', schema) and len(instance) < mP:
        errors.append(ValidationError(
            '{} does not have enough properties'.format(instance)))


def maxProperties(validator, mP, instance, schema, errors):
    if validator.is_type(instance, 'object', schema) and len(instance) > mP:
        errors.append(ValidationError(
            '{} has too many properties'.format(instance)))


def enum(validator, enum, instance, schema, errors):
    if instance not in enum:
        errors.append(ValidationError(
            '{} is not one of {}'.format(instance, enum)))


def uniqueItems(validator, uI, instance, schema, errors):
    if (uI and validator.is_type(instance, 'array', schema) and
            not _utils.is_uniq(instance)):
        errors.append(ValidationError(
            '{} has non-unique elements'.format(instance)))


def not_(validator, not_, instance, schema, errors):
    tmp_errors = []
    validator._descend(instance, not_, tmp_errors)
    if len(tmp_errors) != 0:
        return
    errors.append(ValidationError(
        '{} is not allowed for {}'.format(not_, instance)))


def dependencies(validator, deps, instance, schema, errors):
    raise NotImplementedError


def multipleOf(validator, mO, instance, schema, errors):
    if not validator.is_type(instance, 'number', schema):
        return
    if isinstance(mO, float):
        quotient = instance / mO
        failed = int(quotient) != quotient
    else:
        failed = instance % mO
    if failed:
        errors.append(ValidationError(
            '{} is not a multiple of {}'.format(instance, mO)))


def pattern(validator, pattern, instance, schema, errors):
    if (validator.is_type(instance, 'string', schema) and
            not validator._get_regex(pattern).search(instance)):
        errors.append(ValidationError(
            '{} does not match {}'.format(instance, pattern)))


def patternProperties(validator, pP, instance, schema, errors):
    if not validator.is_type(instance, 'object', schema):
        return
    new_instance = dict(instance)
    for pattern, subschema in pP.items():
        x = validator._get_regex(pattern)
        for k, v in instance.items():
            if x.search(k):
                new_instance[k] = validator._descend(
                    v, subschema, errors, path=k, schema_path=pattern)
    return new_instance, pP



# -*- coding: utf-8 -*-
from urllib.parse import parse_qsl


def _is_primitive_type(type_name):
    return (
        type_name == 'string' or
        type_name == 'integer' or
        type_name == 'number' or
        type_name == 'boolean')


def _convert_style(style, explode, typ, value):
    if _is_primitive_type(typ):
        typ = 'string'
    mapping = {
        ('simple', 'string'): 0,
        ('simple', 'array'): 1,
        ('simple', 'object', True): 2,
        ('simple', 'object', False): 3,
        ('form', 'string'): 0,
        ('form', 'array', True): 0,
        ('form', 'array', False): 1,
        ('form', 'object', False): 3,
    }
    method = mapping.get((style, typ, explode))
    if method is None:
        method = mapping.get((style, typ))
        if method is None:
            raise NotImplementedError  # pragma: no cover
    if method == 0:
        return value
    elif method == 1:
        return value.split(',')
    elif method == 2:
        tmp = value.split(',')
        value = {}
        for t in tmp:
            k, v = parse_qsl(
                t, keep_blank_values=True, strict_parsing=True)[0]
            value[k] = v
        return value
    elif method == 3:
        tmp = value.split(',')
        value = {}
        for i in range(0, len(tmp), 2):
            value[tmp[i]] = tmp[i + 1]
        return value
    raise NotImplementedError('style={}, explode={}, type={}'.format(
        style, explode, typ))  # pragma: no cover


def _convert_type(schema, value):
    typ = schema.get('type', 'object')
    fmt = schema.get('format')
    if typ == 'string':
        return value
    elif typ == 'boolean':
        value = value.lower()
        if value == 'true':
            return True
        elif value == 'false':
            return False
        else:
            raise ValueError('invalid value')
    elif typ == 'integer':
        value = int(value)
        if fmt is None:
            return value
        elif fmt == 'int32':
            rang = (-2**31, 2**31 - 1)
        elif fmt == 'int64':
            rang = (-2**63, 2**63 - 1)
        else:
            raise ValueError('invalid integer format')  # pragma: no cover
        if value < rang[0] or value > rang[1]:
            raise ValueError('out of range')
        return value
    elif typ == 'number':
        return float(value)
    elif typ == 'object':
        props = schema.get('properties', {})
        for pn, ps in props.items():
            if pn in value:
                value[pn] = _convert_type(ps, value[pn])
        aprops = schema.get('additionalProperties', True)
        if isinstance(aprops, dict):
            others = set(value.keys()) - set(props.keys())
            for k in others:
                value[k] = _convert_type(aprops, value[k])
        return value
    elif typ == 'array':
        item_schema = schema['items']
        if isinstance(item_schema, dict):
            for i in range(len(value)):
                value[i] = _convert_type(item_schema, value[i])
        else:
            raise ValueError('array items is must be dict')  # pragma: no cover
        return value
    else:
        raise ValueError('invalid type')  # pragma: no cover



def ensure_list(thing):
    if not isinstance(thing, list):
        return [thing]
    return thing


def merge_instance(base, other):
    if isinstance(base, dict):
        output = dict(base)
        for k, v in other.items():
            if k not in output:
                output[k] = v
            else:
                output[k] = merge_instance(output[k], v)
    elif isinstance(base, list):
        output = list(base)
        for i, v in enumerate(other):
            output[i] = merge_instance(output[i], v)
    else:
        if not isinstance(other, str):
            return other
        return base
    return output


def merge_instances(others):
    if not isinstance(others, list):
        raise ValueError
    base = type(others[0])(others[0])
    for o in others[1:]:
        base = merge_instance(base, o)
    return base


def is_uniq(array):
    def unbool(e, true=object(), false=object()):
        if e is True:
            return true
        elif e is False:
            return false
        return e
    try:
        return len(set(unbool(a) for a in array)) == len(array)
    except Exception:
        seen = []
        for e in array:
            e = unbool(e)
            if e in seen:
                return False
            seen.append(e)
        return True



from base64 import b64decode
import datetime
from ipaddress import IPv4Address, IPv6Address
import re

import rfc3987


RE_HOSTNAME = re.compile(r'^[A-Za-z0-9][A-Za-z0-9\.\-]{1,255}$')


def int32(v):
    if isinstance(v, int) and (v < -2**31 or v > 2**31 - 1):
        raise ValueError('{} is out of int32 range'.format(v))
    return v


def int64(v):
    if isinstance(v, int) and (v < -2**63 or v > 2**63 - 1):
        raise ValueError('{} is out of int64 range'.format(v))
    return v


def byte(v):
    if isinstance(v, str):
        return b64decode(v)
    return v


def date(v):
    if not isinstance(v, str):
        return v
    return datetime.datetime.strptime(v, '%Y-%m-%d').date()


def date_time(v):
    if not isinstance(v, str):
        return v

    v = v.replace(':', '')
    if v[-1] == 'Z':
        v = v[0:-1] + '+0000'
    if '.' in v:
        tries = ['%Y-%m-%dT%H%M%S.%f%z', '%Y-%m-%dT%H%M%S.%f']
    else:
        tries = ['%Y-%m-%dT%H%M%S%z', '%Y-%m-%dT%H%M%S']
    for pattern in tries:
        try:
            ret = datetime.datetime.strptime(v, pattern)
            if ret.tzinfo is None:
                raise ValueError(
                    'invalid date-time format. required timezone info')
            return ret
        except Exception:
            pass
    raise ValueError('invalid date-time format')


def email(v):
    if isinstance(v, str):
        # TODO(kazuki): more strict
        if '@' not in v:
            raise ValueError('invalid email format')
    return v


def hostname(v):
    if isinstance(v, str):
        # TODO(kazuki): more strict
        if not RE_HOSTNAME.match(v):
            raise ValueError('invalid hostname format')
    return v


def ipv4(v):
    if not isinstance(v, str):
        return v
    return IPv4Address(v)


def ipv6(v):
    if not isinstance(v, str):
        return v
    return IPv6Address(v)


def uri(v):
    if isinstance(v, str):
        rfc3987.parse(v, rule='URI')
    return v


def uriref(v):
    if isinstance(v, str):
        rfc3987.parse(v, rule='URI_reference')
    return v



def resolve_refs(node, resolver):
    if isinstance(node, list):
        for v in node:
            resolve_refs(v, resolver)
        return
    if not isinstance(node, dict):
        return
    ref_url = node.get('$ref', None)
    if ref_url is not None:
        node.clear()
        _, fragment = resolver.resolve(ref_url)
        resolve_refs(fragment, resolver)
        node.update(fragment)
        return
    for v in node.values():
        resolve_refs(v, resolver)



from pyramid_oas3.jsonschema import _formats
from pyramid_oas3.jsonschema.exceptions import FormatError


class FormatChecker(object):
    def __init__(self, formats):
        self._formats = formats

    def check(self, instance, format):
        checker = self._formats.get(format)
        if not checker:
            return instance
        try:
            return checker(instance)
        except Exception as e:
            raise FormatError('{} is not a {}'.format(
                instance, format)) from e


class Draft4(FormatChecker):
    def __init__(self):
        super().__init__({
            'date-time': _formats.date_time,
            'email': _formats.email,
            'hostname': _formats.hostname,
            'ipv4': _formats.ipv4,
            'ipv6': _formats.ipv6,
            'uri': _formats.uri,
        })


class Draft5(FormatChecker):
    def __init__(self):
        super().__init__({
            'date-time': _formats.date_time,
            'email': _formats.email,
            'hostname': _formats.hostname,
            'ipv4': _formats.ipv4,
            'ipv6': _formats.ipv6,
            'uri': _formats.uri,
            'uriref': _formats.uriref,
        })


class OAS3(FormatChecker):
    def __init__(self):
        super().__init__({
            'int32': _formats.int32,
            'int64': _formats.int64,
            'byte': _formats.byte,
            'date': _formats.date,
            'date-time': _formats.date_time,
            'email': _formats.email,
            'hostname': _formats.hostname,
            'ipv4': _formats.ipv4,
            'ipv6': _formats.ipv6,
            'uri': _formats.uri,
            'uriref': _formats.uriref,
        })



# -*- coding: utf-8 -*-
import pyramid

from pyramid_oas3.tween import (
    validation_tween_factory, ResponseValidationError)
from pyramid_oas3.jsonschema.exceptions import ValidationErrors

__all__ = [
    'validation_tween_factory',
    'ValidationErrors',
    'ResponseValidationError',
]


def includeme(config):
    config.add_tween(
        "pyramid_oas3.validation_tween_factory",
        under=pyramid.tweens.EXCVIEW
    )



import numbers
from functools import lru_cache
import re

from pyramid_oas3.jsonschema.exceptions import (
    UnknownTypeError, ValidationErrors)
from pyramid_oas3.jsonschema._resolvers import _Resolver
from pyramid_oas3.jsonschema import _validators, formats


class _Validator(object):
    def __init__(
            self, validators, format_checker, schema, resolver=None,
            fill_by_default=False):
        self._validators = validators
        self._schema = schema
        self._fill_by_default = fill_by_default
        self._types = {
            'array': list,
            'boolean': bool,
            'integer': int,
            'null': type(None),
            'number': numbers.Number,
            'object': dict,
            'string': str,
        }
        self.format_checker = format_checker
        self.resolver = resolver
        if resolver is None:
            self.resolver = _Resolver('', schema)
        self._get_regex = lru_cache(8192)(re.compile)

    def _validate(self, instance, schema, errors):
        scope = schema.get('id')
        if scope:
            self.resolver.push_scope(scope)
        try:
            ref = schema.get('$ref')
            if ref is not None:
                schema = {'$ref': ref}
            new_instance = instance
            new_schema = {}
            for k, v in schema.items():
                validator = self._validators.get(k)
                if validator is None:
                    continue
                if ref is None and self._is_nullable_and_skip_validation(
                        instance, schema):
                    continue
                ret = validator(self, v, instance, schema, errors)
                if ret is None:
                    tmp_instance, tmp_schema = instance, v
                else:
                    tmp_instance, tmp_schema = ret
                if id(tmp_instance) != id(instance):
                    new_instance = tmp_instance
                if k == '$ref':
                    new_schema.update(tmp_schema)
                else:
                    new_schema[k] = tmp_schema
            return new_instance, new_schema
        finally:
            if scope:
                self.resolver.pop_scope()

    def _descend(self, instance, schema, errors, path=None, schema_path=None):
        err_off = len(errors)
        ret = self._validate(instance, schema, errors)
        for error in errors[err_off:]:
            if path is not None:
                error.path.appendleft(path)
            if schema_path is not None:
                error.schema_path.appendleft(schema_path)
        return ret

    def validate(self, instance, **kwargs):
        errors = []
        new, schema = self._validate(instance, self._schema, errors, **kwargs)
        if errors:
            raise ValidationErrors(errors)
        return new, schema

    def is_valid(self, instance, **kwargs):
        try:
            self.validate(instance)
            return True
        except NotImplementedError as e:
            raise
        except Exception:
            return False

    def is_type(self, instance, typ, schema):
        pytype = self._types.get(typ)
        if pytype is None:
            raise UnknownTypeError(typ, instance, schema)
        if isinstance(instance, bool) and issubclass(int, pytype):
            return False
        return isinstance(instance, pytype)

    def _is_nullable_and_skip_validation(self, instance, schema):
        if instance is not None:
            return False
        return schema.get('nullable', False)


class Draft4(_Validator):
    def __init__(self, schema, format_checker=None, **kwargs):
        if format_checker is None:
            format_checker = formats.Draft4()
        super().__init__({
            '$ref': _validators.ref,
            'type': _validators.type,
            'items': _validators.items,
            'properties': _validators.properties,
            'required': _validators.required,
            'format': _validators.format,
            'oneOf': _validators.oneOf,
            'additionalItems': _validators.additionalItems,
            'additionalProperties': _validators.additionalProperties,
            'allOf': _validators.allOf,
            'anyOf': _validators.anyOf,
            'minItems': _validators.minItems,
            'maxItems': _validators.maxItems,
            'minProperties': _validators.minProperties,
            'maxProperties': _validators.maxProperties,
            'minLength': _validators.minLength,
            'maxLength': _validators.maxLength,
            'minimum': _validators.minimum,
            'maximum': _validators.maximum,
            'enum': _validators.enum,
            'uniqueItems': _validators.uniqueItems,
            'not': _validators.not_,
            'dependencies': _validators.dependencies,
            'multipleOf': _validators.multipleOf,
            'pattern': _validators.pattern,
            'patternProperties': _validators.patternProperties,
        }, format_checker, schema, **kwargs)


class Draft5(Draft4):
    def __init__(self, schema, format_checker=None, **kwargs):
        if format_checker is None:
            format_checker = formats.Draft5()
        super().__init__(schema, format_checker=format_checker, **kwargs)


class OAS3(Draft5):
    def __init__(self, schema, format_checker=None, **kwargs):
        if format_checker is None:
            format_checker = formats.OAS3()
        super().__init__(schema, format_checker=format_checker, **kwargs)



from collections import deque
import pprint
import textwrap


def _dedent(x):
    return textwrap.dedent(x).strip('\n')


def _pformat_and_indent(obj, times=1):
    return textwrap.indent(pprint.pformat(obj), ' ' * (4 * times))


def _format_as_index(indices):
    if not indices:
        return ''
    return '[{}]'.format(']['.join(repr(index) for index in indices))


_unset = object()


class UnknownTypeError(Exception):
    def __init__(self, type, instance, schema):
        self.type = type
        self.instance = instance
        self.schema = schema

    def __str__(self):
        return _dedent('''
        Unknown type "{}" for validator with schema:
        {}

        While checking instance:
        {}''').format(
            self.type,
            _pformat_and_indent(self.schema),
            _pformat_and_indent(self.instance))


class _Error(Exception):
    def __init__(
            self, message, validator=_unset, path=(), cause=None, context=(),
            validator_value=_unset, instance=_unset, schema=_unset,
            schema_path=(), parent=None):
        self.message = message
        self.path = self.relative_path = deque(path)
        self.schema_path = self.relative_schema_path = deque(schema_path)
        self.context = list(context)
        self.__cause__ = cause
        self.validator = validator
        self.validator_value = validator_value
        self.instance = instance
        self.schema = schema
        self.parent = parent

        for error in context:
            error.parent = self

    def __repr__(self):
        return "<%s: %r>" % (self.__class__.__name__, self.message)

    def __str__(self):
        essential_for_verbose = (
            self.validator, self.validator_value, self.instance, self.schema,
        )
        if any(m is _unset for m in essential_for_verbose):
            return self.message

        return self.message + _dedent(
            '''
            Failed validating %r in %s%s:
            %s

            On %s%s:
            %s
            ''').format(
                self.validator,
                self._word_for_schema_in_error_message,
                _format_as_index(list(self.relative_schema_path)[:-1]),
                _pformat_and_indent(self.schema),
                self._word_for_instance_in_error_message,
                _format_as_index(self.relative_path),
                _pformat_and_indent(self.instance),
            )

    @property
    def absolute_path(self):
        parent = self.parent
        if parent is None:
            return self.relative_path

        path = deque(self.relative_path)
        path.extendleft(reversed(parent.absolute_path))
        return path

    @property
    def absolute_schema_path(self):
        parent = self.parent
        if parent is None:
            return self.relative_schema_path

        path = deque(self.relative_schema_path)
        path.extendleft(reversed(parent.absolute_schema_path))
        return path


class ValidationError(_Error):
    _word_for_schema_in_error_message = "schema"
    _word_for_instance_in_error_message = "instance"


class SchemaError(_Error):
    _word_for_schema_in_error_message = "metaschema"
    _word_for_instance_in_error_message = "schema"


class RefResolutionError(Exception):
    pass


class FormatError(Exception):
    def __init__(self, message, cause=None):
        self.message = message
        if cause:
            self.__cause__ = cause

    def __str__(self):
        return self.message


class StyleError(FormatError):
    pass


class ValidationErrors(Exception):
    def __init__(self, errors):
        if isinstance(errors, Exception):
            errors = [errors]
        self.errors = errors

    def __str__(self):
        return 'found {} validation error(s)\n{}'.format(
            len(self.errors), '\n\n'.join([str(e) for e in self.errors]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import setup


def _load_lines(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return [l.strip() for l in f.readlines()]


setup(
    name='pyramid_oas3',
    version='0.1.1',
    description='OpenAPI 3.0 Validator for Pyramid',
    packages=['pyramid_oas3', 'pyramid_oas3.jsonschema'],
    author='Kazuki Oikawa',
    author_email='k@oikw.org',
    license='MIT',
    url='http://github.com/kazuki/pyramid-oas3',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=_load_lines('requirements.txt'),
    tests_requires=_load_lines('test-requirements.txt'),
    test_suite='nose2.collector.collector',
)



# -*- coding: utf-8 -*-
import json
from urllib.parse import parse_qs, urlparse

from pyramid.httpexceptions import (
    HTTPBadRequest, HTTPNotAcceptable, HTTPUnauthorized)

from pyramid_oas3.jsonschema import OAS3Validator, Resolver
from pyramid_oas3.jsonschema.exceptions import (
    ValidationErrors, ValidationError, StyleError)
from pyramid_oas3.resolve import resolve_refs
from pyramid_oas3.types import _convert_type, _convert_style


MIME_JSON = 'application/json'


def validation_tween_factory(handler, registry):
    from pyramid.interfaces import IRoutesMapper
    settings = {
        k[13:]: v
        for k, v in registry.settings.items()
        if k.startswith('pyramid_oas3.')}
    validate_response = settings.get('validate_response', False)
    fill_default = settings.get('fill_by_default', False)
    schema = settings['schema']
    resolver = Resolver('', schema)
    default_security = schema.get('security', None)
    route_mapper = registry.queryUtility(IRoutesMapper)
    prefixes = list(set([
        urlparse(server['url']).path.rstrip('/')
        for server in schema.get('servers', [])]))
    if not prefixes:
        prefixes = ['/']

    # 検証を簡単にするためにパラメータの $ref をすべて解決する。
    paths = schema['paths']
    for path, path_item in paths.items():
        for method, op_obj in path_item.items():
            if method not in ('get', 'put', 'post', 'delete', 'options',
                              'head', 'patch', 'trace'):
                continue
            params = op_obj.get('parameters')
            if params:
                resolve_refs(params, resolver)

    def Validator(schema):
        return OAS3Validator(
            schema, resolver=resolver, fill_by_default=fill_default)

    def get_operation_object(path, method):
        if path and path[0] != '/':
            path = '/' + path
        for prefix in prefixes:
            if not path.startswith(prefix):
                continue
            path_item = paths.get(path[len(prefix):])
            if path_item:
                return path_item.get(method.lower())
        return None

    def validator_tween(request):
        route_info = route_mapper(request)
        route = route_info.get('route', None)
        if not route:  # pragma: no cover
            return handler(request)
        op_obj = get_operation_object(route.path, request.method)
        if not op_obj:  # pragma: no cover
            return handler(request)

        _check_security(default_security, request, op_obj)
        params, body = _validate_and_parse(
            Validator, resolver, request, route_info.get('match', {}),
            op_obj, fill_default)

        def oas3_data(_):
            return params

        def oas3_body(_):
            return body

        request.set_property(oas3_data)
        request.set_property(oas3_body)
        response = handler(request)
        if not validate_response:
            return response

        try:
            responses_obj = op_obj['responses']
            res_obj = responses_obj.get(
                str(response.status_code), responses_obj.get('default'))
            if res_obj is None:
                raise ValidationErrors(ValidationError(
                    'invalid response status code'))
            content_prop = res_obj.get('content')
            has_content = content_prop and len(content_prop.keys()) > 0
            if not has_content and response.has_body:
                raise ValidationErrors(ValidationError(
                    'invalid response: body must be empty'))
            if response.content_type == MIME_JSON:
                res_schema = content_prop.get(MIME_JSON, {}).get('schema')
                if res_schema:
                    res_json = json.loads(response.body.decode('utf8'))
                    _validate(Validator, res_schema, res_json)
        except Exception as e:
            raise ResponseValidationError(response, e)
        return response
    return validator_tween


def _check_security(default_security, request, op_obj):
    requires = op_obj.get('security', default_security)
    if not requires:
        return
    if request.authenticated_userid is None:
        raise HTTPUnauthorized


def _validate_and_parse(
        Validator, resolver, request, path_matches, op_obj, fill_by_default):
    params, queries = {}, {}
    if request.query_string:
        try:
            queries = parse_qs(request.query_string,
                               keep_blank_values=True, strict_parsing=True)
        except Exception:
            raise HTTPBadRequest('cannot parse query string')
    for param_obj in op_obj.get('parameters', []):
        params.update(_validate_and_parse_param(
            Validator, request, param_obj, path_matches, queries,
            fill_by_default))

    body = None
    reqbody = op_obj.get('requestBody')
    if reqbody:
        if '$ref' in reqbody:
            _, reqbody = resolver.resolve(reqbody['$ref'])
        accept_types = set(reqbody.get('content', {}).keys())
        if not accept_types or request.content_type not in accept_types:
            raise HTTPNotAcceptable
        media_type_obj = reqbody.get('content', {}).get(MIME_JSON)
        if media_type_obj is not None:
            required = reqbody.get('required', False)
            body = request.body
            if required and not body:
                raise ValidationErrors(ValidationError(
                    'json body is required'))
            body = json.loads(body.decode('utf8'))
            json_schema = media_type_obj.get('schema')
            if json_schema:
                body = _validate(Validator, json_schema, body)
    return params, body


def _validate_and_parse_param(
        Validator, request, param_obj, path_matches, queries, fill_by_default):
    if param_obj.get('allowEmptyValue', False):
        raise NotImplementedError  # pragma: no cover
    in_, name, schema, value = (
        param_obj['in'], param_obj['name'], param_obj.get('schema'), None)
    style = param_obj.get(
        'style', 'form' if in_ in ('query', 'cookie') else 'simple')
    explode = param_obj.get('explode', True if style == 'form' else False)
    if not schema:
        typ = 'object'
    elif 'type' in schema:
        typ = schema['type']
    else:
        # QUICKHACK(kazuki): 同じプリミティブ型のoneOf/anyOfのみ許容する
        def _check_subschemas(s, t):
            st = s.get('type')
            if st is not None:
                if t is None:
                    return st
                if t == st:
                    return t
                raise NotImplementedError
            ss_list = s.get('oneOf', s.get('anyOf'))
            if ss_list is None:
                raise NotImplementedError
            for ss in ss_list:
                t = _check_subschemas(ss, t)
            return t
        typ = _check_subschemas(schema, None)
        if typ is None:
            raise NotImplementedError

    if style in ('matrix', 'label', 'spaceDelimited', 'pipeDelimited'):
        raise NotImplementedError  # pragma: no cover

    if in_ == 'path':
        value = path_matches[name]  # always successful
    elif in_ == 'query':
        if style == 'form' and typ == 'object' and explode:
            raise NotImplementedError  # pragma: no cover
        if style == 'deepObject':
            value = {}
            for k, v in queries.items():
                if k.startswith(name + '[') and k[-1] == ']':
                    value[k[len(name)+1:-1]] = v[0]
            if not value:
                value = None
        else:
            if name in queries:
                value = queries[name]
                if not (style == 'form' and explode and typ == 'array'):
                    value = value[0]
            elif schema and fill_by_default and 'default' in schema:
                return {name: schema['default']}
    elif in_ == 'header' and name in request.headers:
        value = request.headers[name]
    elif in_ == 'cookie':  # pragma: no cover
        raise NotImplementedError
    if param_obj.get('required', False) and value is None:
        raise ValidationErrors(ValidationError(
            'required parameter "{}" is not found in {}'.format(name, in_)))
    if value is None:
        return {}

    if not isinstance(value, dict):
        try:
            value = _convert_style(style, explode, typ, value)
        except Exception as e:
            raise ValidationErrors(StyleError(
                'invalid style of "{}": {}'.format(name, e)))
    if schema:
        try:
            value = _convert_type(schema, value)
        except Exception as e:
            raise ValidationErrors(ValueError(
                'invalid value of "{}": {}'.format(name, e)))
        value = _validate(Validator, schema, value)
    return {name: value}


def _validate(Validator, schema, instance):
    validator = Validator(schema)
    new_instance, _ = validator.validate(instance)
    return new_instance


class ResponseValidationError(Exception):
    def __init__(self, response, exception):
        self.response = response
        self.exception = exception



from collections.abc import MutableMapping, Sequence
from functools import lru_cache
from urllib.parse import urljoin, urlsplit, urlunsplit, unquote


class _Resolver(object):
    def __init__(self, base_uri, referrer):
        self._scopes_stack = [base_uri]
        self._store = _URIDict()
        self._store[base_uri] = referrer
        self._urljoin = lru_cache(64)(urljoin)
        self._remote_cache = lru_cache(8192)(self.resolve_from_url)

    @property
    def resolution_scope(self):
        return self._scopes_stack[-1]

    def push_scope(self, scope):
        self._scopes_stack.append(self._urljoin(self.resolution_scope, scope))

    def pop_scope(self):
        try:
            self._scopes_stack.pop()
        except IndexError:
            raise RuntimeError

    def resolve(self, ref):
        if self.resolution_scope:
            url = self._urljoin(self.resolution_scope, ref)
        else:
            url = ref
        return url, self._remote_cache(url)

    def resolve_from_url(self, url):
        url, fragment = _urldefrag(url)
        document = self._store.get(url)
        if document is None:
            raise NotImplementedError
        return self.resolve_fragment(document, fragment)

    def resolve_fragment(self, document, fragment):
        fragment = fragment.lstrip('/')
        parts = unquote(fragment).split('/') if fragment else []
        for part in parts:
            part = part.replace('~1', '/').replace('~0', '~')
            if isinstance(document, Sequence):
                try:
                    part = int(part)
                except ValueError:
                    pass
            try:
                document = document[part]
            except (TypeError, LookupError):
                raise RuntimeError(
                    "Unresolvable JSON pointer: {}".format(fragment)
                )
        return document


class _URIDict(MutableMapping):
    def _normalize(self, uri):
        return urlsplit(uri).geturl()

    def __init__(self, *args, **kwargs):
        self._store = dict()
        self._store.update(*args, **kwargs)

    def __getitem__(self, uri):
        return self._store[self._normalize(uri)]

    def __setitem__(self, uri, value):
        self._store[self._normalize(uri)] = value

    def __delitem__(self, uri):
        del self._store[self._normalize(uri)]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)


def _urldefrag(url):
    if '#' in url:
        s, n, p, q, frag = urlsplit(url)
        defrag = urlunsplit((s, n, p, q, ''))
    else:
        defrag = url
        frag = ''
    return defrag, frag
